import asyncio

from asyevent.callback import Callback
from asyevent.event import Event
from asyevent.command import Command

from asyevent.exceptions import (
    CommandNotFound,
    EventNotFound,
    CommandAlreadyRegistered,
    EventAlreadyRegistered,
)

from typing import Callable, Optional, Union, List, Tuple


class EventManager:
    """
    The manager stores events and commands, it
    provides an error handler for them.
    """

    def __init__(self):
        """
        Initialises an event manager.
        """
        self._events: List[Event] = []
        self._commands: List[Command] = []

        self.loop = asyncio.get_event_loop()

        # For all events and commands that define `handle_errors` to True, their
        # exceptions are handled in this event.
        # Passed parameters are : the exception, the event, the callback, args, **kwargs.
        self.error_handler = self.create_event("<error_handler>", handle_errors=False)

    @property
    def events(self) -> Tuple[Event]:
        """
        Returns all registered events.

        :return: A tuple of events.
        """
        return tuple(self._events)

    @property
    def commands(self) -> Tuple[Command]:
        """
        Returns all registered commands.

        :return: A tuple of commands.
        """
        return tuple(self._commands)

    def as_command(
        self,
        *,
        name: str = None,
        handle_errors: bool = True,
        multiple_callbacks: bool = True,
        priority: int = 1,
        **options
    ) -> Callable[[Union[Callable, Callback]], Callback]:
        """
        A decorator which registers a coroutine as a callback of this command.

        :param priority: When an event is raised, associated callbacks will be invoked by ascending priority order.
        :param name: The command name. Coroutine's name by default.
        :param handle_errors: Determines if exceptions are handle into the error handler event.
        :param multiple_callbacks: Does the command allow multiple callbacks.

        :return: A callback.
        """

        def decorator(callback: Union[Callable, Callback]) -> Callback:
            command = self.create_command(
                callback,
                name=name,
                handle_errors=handle_errors,
                multiple_callbacks=multiple_callbacks,
                priority=priority,
                **options
            )

            return command.initial_callback

        return decorator

    def create_command(
        self,
        callback: Union[Callable, Callback],
        *,
        name: str = None,
        handle_errors: bool = True,
        multiple_callbacks: bool = True,
        priority: int = 1,
        **options
    ) -> Command:
        """
        Registers a coroutine as a `command` event.
        The purpose of using this instead of `Event.as_callback` is dynamic event names.

        :param multiple_callbacks: Does the command allow multiple callbacks associated to.
        :param callback: Coroutine which will be called when the command event is raised.
        :param name: The command name. Coroutine name by default.
        :param handle_errors: Determines if exceptions are handle into the error handler event.
        :param priority: When an event is raised, associated callbacks will be invoked by ascending priority order.
        :param options: Callback options

        :return: A command.
        """
        command = Command(
            callback,
            event_manager=self,
            name=name,
            handle_errors=handle_errors,
            multiple_callbacks=multiple_callbacks,
            priority=priority,
            **options
        )
        self._commands.append(command)

        return command

    def create_event(
        self, name: str, *, handle_errors: bool = True, multiple_callbacks: bool = True
    ) -> Event:
        """
        Creates an event associated to the manager.

        :param multiple_callbacks: Does the command allow multiple callbacks associated to.
        :param name: The event name.
        :param handle_errors: Determines if exceptions are handle into the error handler event.

        :return: The event.
        """
        event = Event(
            name=name,
            event_manager=self,
            handle_errors=handle_errors,
            multiple_callbacks=multiple_callbacks,
        )
        self._events.append(event)

        return event

    def get_command(
        self, name: str, *, case_sensitive: bool = True
    ) -> Optional[Command]:
        """
        Get a command by its name. The command must be associate to this manager.

        :param name: The command name.
        :param case_sensitive: Is case sensitive.

        :return: A command if found.
        """

        def checker(cmd: Command) -> bool:
            if case_sensitive:
                return cmd.command_name == name

            return cmd.command_name.casefold() == name.casefold()

        return next(filter(checker, self._commands), None)

    def get_event(self, name: str, *, case_sensitive: bool = True) -> Optional[Event]:
        """
        Get an event by its name. The event must be associate to this manager.

        :param name: The event name.
        :param case_sensitive: Is case sensitive.

        :return: An event if found.
        """

        def checker(event: Event) -> bool:
            if case_sensitive:
                return event.event_name == name

            return event.event_name.casefold() == name.casefold()

        return next(filter(checker, self._events), None)

    def replace_command_name(self, name: str, *, new_name: str):
        """
        Replace a command name.

        :param name: The name of the command to replace.
        :param new_name: The new name.

        :raise: CommandNotFound: If no command was found.
        """
        command = self.get_command(name)

        if command is None:
            raise CommandNotFound(name=name)

        command.command_name = new_name

    def add_event(self, event: Event):
        """
        Add an event to this event manager, and remove it from the previous one.

        :param event: The event to add.

        :raise: EventAlreadyRegistered: If the event is already registered.
        """
        if event in self._events:
            raise EventAlreadyRegistered(event=event)

        event.event_manager.remove_event(event)
        self._events.append(event)
        event.event_manager = self

    def remove_event(self, event: Event):
        """
        Remove an event from this event manager.

        :param event: The event to remove.
        """
        self._events.remove(event)
        event.event_manager = None

    def remove_event_by_name(self, name: str):
        """
        Remove an event by its name from this event manager.

        :param name: The name of the event to remove.

        :raise: EventNotFound: If no event was found.
        """
        event = self.get_event(name)

        if event is None:
            raise EventNotFound(name=name)

        self.remove_event(event)

    def add_command(self, command: Command):
        """
        Add a command to this event manager, and remove it from the previous one.

        :param command: The command to add.

        :raise: CommandAlreadyRegistered: If the command is already registered.
        """
        if command in self._commands:
            raise CommandAlreadyRegistered(command=command)

        command.event_manager.remove_command(command)
        self._commands.append(command)
        command.event_manager = self

    def remove_command(self, command: Command):
        """
        Remove a command from this event manager.

        :param command: The command to remove.
        """
        self._commands.remove(command)
        command.event_manager = None

    def remove_command_by_name(self, name: str):
        """
        Remove a command by its name from this event manager.

        :param name: The name of the command to remove.

        :raise: CommandNotFound: If no command was found.
        """
        command = self.get_command(name)

        if command is None:
            raise CommandNotFound(name=name)

        self.remove_command(command)

    async def invoke_command(
        self, _name: str, *args, _case_sensitive: bool = True, **kwargs
    ):
        """
        Invokes a command by its name.

        :param _name: The command name.
        :param args: The command parameters.
        :param _case_sensitive: Is event research case sensitive.
        :param kwargs: The command keyword parameters.

        :raise CommandNotFound: If no command was found.:
        """
        command = self.get_command(_name, case_sensitive=_case_sensitive)

        if command is None:
            raise CommandNotFound(_name)

        await command.raise_event(*args, **kwargs)

    async def raise_event(
        self, _name: str, *args, _case_sensitive: bool = True, **kwargs
    ):
        """
        Raises an event by its name.

        :param _name: The event name.
        :param _case_sensitive: Is event research case sensitive.
        :param args: The event parameters.
        :param kwargs: The event keyword parameters.

        :raise EventNotFound: If no event was found.:
        """
        event = self.get_event(_name, case_sensitive=_case_sensitive)

        if event is None:
            raise EventNotFound(_name)

        await event.raise_event(*args, **kwargs)



"""
All specific package's exceptions.
Cannot type hint custom classes due to circular import.
"""

from typing import Any


class AsyeventException(Exception):
    """
    Base exception.
    """

    pass


class EventException(AsyeventException):
    """
    Base event exception.
    """

    def __init__(self, message: str, event=None):
        self.event = event
        super().__init__(message)


class CommandException(AsyeventException):
    """
    Base command exception.
    """

    def __init__(self, message: str, command=None):
        self.command = command
        super().__init__(message)


class EventNotFound(EventException):
    """
    Raised when a unknown event is invoked.
    """

    def __init__(self, name: str):
        self.name = name
        super().__init__(f"Event {self.name!r} cannot be found.")


class EventAlreadyExists(EventException):
    """
    Raised when an event that is already defined by its event_name is created.
    """

    def __init__(self, event):
        super().__init__(f"Event {event.event_name!r} already exists.", event)


class EventAlreadyRegistered(EventException):
    """
    Raised when an event is add to en event_manager that already contains the event.
    """

    def __init__(self, event):
        super().__init__(f"Event {event.event_name!r} is already registered.", event)


class CommandNotFound(CommandException):
    """
    Raised when a unknown command is invoked.
    """

    def __init__(self, name: str):
        self.name = name
        super().__init__(f"Command {self.name!r} cannot be found.")


class CommandAlreadyExists(CommandException):
    """
    Raised when an event that is already defined by its event_name is created.
    """

    def __init__(self, command):
        super().__init__(f"Command {command.command_name!r} already exists.", command)


class CommandAlreadyRegistered(CommandException):
    """
    Raised when a command is add to en event_manager that already contains the command.
    """

    def __init__(self, command):
        super().__init__(
            f"Command {command.command_name!r} is already registered.", command
        )


class ParsingError(AsyeventException):
    """
    Raised on function parameters parsing failed.
    """

    def __init__(self, value: Any, excepted_type: type):
        self.value = value
        self.excepted_type = excepted_type
        super().__init__(
            f"Parameter {value!r} of type {type(value)} cannot be parsed into {excepted_type}."
        )


class ParsingNotImplemented(AsyeventException):
    """
    Raised when trying to parse a parameter into a no `IParsable` class.
    """

    def __init__(self, value: Any, excepted_type: type):
        self.message = (
            f"Parameter {value!r} is not of type {excepted_type}, but {excepted_type} is not parsable. "
            f"\nImplement `IParsable` to make a class parsable."
        )
        super().__init__(self.message)



from __future__ import annotations

import asyncio
import time
import typing

from asyevent.callback import Callback
from asyevent.exceptions import EventAlreadyExists


class Event:
    """
    Do not manually initialise this class. Use instead `EventManager.create_event()`.

    Events contain `Callback` objects. They can be raised for invoke all callbacks stored.
    """

    def __init__(
        self,
        name: str,
        *,
        event_manager,
        handle_errors: bool = True,
        multiple_callbacks: bool = True,
    ):
        """
        Initialises an event.

        :param name: The event name. Must be unique in the event manager.
        :param event_manager: An event manager.
        :param handle_errors: Are the errors handled into the `event_manager` error handler.
        :param multiple_callbacks: Does the command allow multiple callbacks associated to.

        :raise EventAlreadyExists: If the event name is not unique in the event_manager.
        """
        if event_manager.get_event(name, case_sensitive=False):
            raise EventAlreadyExists(event_manager.get_event(name))

        self.event_name = name

        self.handle_errors = handle_errors
        self.event_manager = event_manager
        self.multiple_callbacks = multiple_callbacks

        self.pass_extra_after = False

        self._before = None
        self._after = None

        self._loop = asyncio.get_event_loop()
        self._tasks: typing.List[asyncio.Task] = []
        self._callbacks: typing.Dict[int, typing.List[Callback]] = {}

        self._internal_event = asyncio.Event()

    def __iadd__(self, callback: Callback):
        # default priority, use `.add_callback()` to custom it.
        self.add_callback(callback)

        return self

    def __isub__(self, callback: Callback):
        self.remove_callback(callback)

        return self

    def __await__(self) -> typing.Generator[typing.Any, None, typing.Any]:
        return self._internal_event.wait().__await__()

    async def __call__(self, *args, **kwargs):
        await self.raise_event(*args, **kwargs)

    def before(self, *, handle_errors: bool = True) -> Event:
        """
        Returns an event that is raised before callbacks.
        The current parameters will be passed to the callback.

        :param handle_errors: Are the errors handled into the `event_manager` error handler.
        """
        if self._before is None:
            self._before = self.event_manager.create_event(
                f"<before:{self.event_name}>", handle_errors=handle_errors
            )

        return self._before

    def after(self, *, handle_errors: bool = True, pass_extra: bool = False) -> Event:
        """
        Returns an event that is raised after the callbacks.
        Their execution duration in seconds as
        well as the current parameters, will be passed to the callback.

        :param handle_errors: Are the errors handled into the `event_manager` error handler.
        :param pass_extra: Pass the callbacks execution time if it is set to `True`.

        """
        self.pass_extra_after = pass_extra

        if self._after is None:
            self._after = self.event_manager.create_event(
                f"<after:{self.event_name}>", handle_errors=handle_errors
            )

        return self._after

    @property
    def callbacks(self) -> typing.Tuple[Callback]:
        """
        Sorts callbacks by layer level desc and return them into a tuple.

        :return: A tuple of callbacks.
        """

        return tuple(
            callback
            for key in reversed(sorted(self._callbacks))
            for callback in self._callbacks[key]
        )

    def as_callback(
        self, *, priority: int = 1, **options
    ) -> typing.Callable[[typing.Union[typing.Callable, Callback]], Callback]:
        """
        A decorator which registers a coroutine as a callback of this event.
        All event's callbacks are invoked when `raise_event` is called.

        :return: `Callback` object.
        """

        def decorator(callback: typing.Union[typing.Callable, Callback]) -> Callback:
            return self.create_callback(callback, priority=priority, **options)

        return decorator

    def create_callback(
        self,
        callback: typing.Union[typing.Callable, Callback],
        *,
        priority: int = 1,
        **options,
    ) -> Callback:
        """
        Creates a callback and add it to this event.

        :param callback: The coroutine that will be executed
        :param priority: When the event is raised, callbacks are invoked in priority ascending order.
        :param options: Callback options.

        :return: The created callback.
        """
        callback = Callback(callback, **options)

        self.add_callback(callback, priority=priority)

        return callback

    def add_callback(self, callback: Callback, *, priority: int = 1):
        """
        Registers a callback to this event.

        :param callback: The callback to register.
        :param priority: When the event is raised, callbacks are invoked in priority ascending order.

        :raise ValueError: If the callback is already registered or if there are multiple callbacks and `multiple_callbacks` is `False`.:
        """
        if not self.multiple_callbacks and self.callbacks:
            raise ValueError(
                f"Cannot add multiple callbacks on event {self.event_name!r}."
            )

        if callback in self.callbacks:
            raise ValueError(f"Callback {callback.__name__!r} is already registered.")

        if self._callbacks.get(priority) is None:
            self._callbacks[priority] = []

        self._callbacks[priority].append(callback)

    def remove_callback(self, callback: Callback):
        """
        Removes a callback from this event.

        :param callback: The callback to remove.

        :raise ValueError: If the callbacks is not registered in this event.:
        """
        new = {
            p: [c_ for c_ in c if c_ != callback] for p, c in self._callbacks.items()
        }

        if new == self._callbacks:
            raise ValueError(f"Callback {callback.__name__!r} is not registered.")

        self._callbacks = new

    async def raise_event(self, *args, **kwargs):
        """
        Calls all registered callbacks. If `handle_errors` is True, `error_handler` event will be raised.
        Parameters passed are : :class: `Exception` the error, :class: `Event` the event where the error occurred,
        *args and **kwargs passed into the initial event.

        :param args: Invocation parameters.
        :param kwargs: Invocation keyword parameters.
        """
        if self._before is not None and self._before.callbacks:
            await self._before.raise_event(*args, **kwargs)

        self._internal_event.set()
        tasks: typing.List[asyncio.Task] = []
        start_time = time.time()

        for callback in self.callbacks:
            task = self._loop.create_task(
                callback.invoke(
                    *args,
                    **kwargs,
                    _event=self,
                    _handler=self.event_manager.error_handler
                    if self.handle_errors
                    else None,
                )
            )

            # appends task to the global view
            self._tasks.append(task)
            # appends task to the local view
            tasks.append(task)

        # waits for local tasks to complete
        await asyncio.gather(*tasks)

        # clears local tasks
        self._tasks = [t for t in self._tasks if t not in tasks]

        if self._after is not None and self._after.callbacks:
            # pass extra parameter only if specified
            if self.pass_extra_after:
                args = (time.time() - start_time, *args)

            await self._after.raise_event(*args, **kwargs)

    def cancel(self):
        """
        Cancel all callbacks that run.
        """
        for task in self._tasks:
            task.cancel()
            self._tasks.remove(task)



from asyevent.callback import Callback

from typing import List


class EventWrapper:
    """
    !! Implement this class to create callbacks from classmethods !!

    Implements a way to deal with classmethod coroutines.
    """

    def __init__(self):
        self.callbacks: List[Callback] = []

        # initialises callbacks
        self._init_callbacks()

    def _init_callbacks(self):
        """
        Pass to all contained callbacks a `self instance` parameter.
        Use this if a callback has been added after the instance initialisation.

        :raise TypeError: If a registered callback is not a classmethod.
        """

        for attr in dir(self):
            if attr != self._init_callbacks.__name__:  # avoid recursive calls
                if isinstance(getattr(self, attr), Callback):
                    self.callbacks.append(getattr(self, attr))

        for callback in self.callbacks:
            if not callback.is_classmethod:
                raise TypeError("Wrapper's callbacks must be as class callbacks.")

            callback.wrapper = self

    def load(self):
        """
        Enables all callbacks contained in implemented class.
        """
        for callback in self.callbacks:
            callback.enable()

    def unload(self):
        """
        Disables all callbacks contained in implemented class.
        """
        for callback in self.callbacks:
            callback.disable()



import typing


def checker(lamb: typing.Callable[..., bool], *, error: Exception):
    def wrapper(*args, **kwargs) -> bool:
        if lamb(*args, **kwargs):
            return True

        raise error

    return wrapper



from __future__ import annotations

import asyncio
import random

from typing import Union
from warnings import warn

# imports needed classes
from asyevent import Event, EventManager, EventWrapper, Callback


# creates an event manager
manager = EventManager()

# creates a global event
start_event = manager.create_event("starting")


class Data:
    """
    Creating our own parsing type by implementing IParable.
    """

    def __init__(self, value: Union[int, float, str]):
        self.value = value

    def __repr__(self) -> str:
        """
        Implement a representation.
        """
        return f">> Data :: {self.value!r} <<"

    @classmethod
    def __parse__(cls, value: Union[int, float, str]) -> Data:
        """
        The method that returns an `Data` instance from a value.
        """
        return cls(value)


class Server(EventWrapper):
    """
    An example server class which is listening for data.
    It inherits from `EventWrapper` in order to use `self` param.
    """

    def __init__(self, m: EventManager):
        super().__init__()

        # creates events
        self.data_received = m.create_event("data_received")
        self.data_lost = m.create_event("data_lost")
        self.data_ok = m.create_event("data_ok")

        # handle_errors set to false avoid exceptions to be catch and pass
        self.shutdown = m.create_event("shut_down", handle_errors=False)

    # `start_delay` is the time took the the callback to start perform
    @start_event.as_callback(start_delay=1)
    async def process_data(self, data: str):
        """
        An example method which processes data
        """
        print("\n---- starting process ----\n")

        # raises the data received event if the data correctly processed
        # pass `data` as a parameter of the event
        await self.data_received.raise_event(data)

        # raises the shutdown event
        await self.shutdown.raise_event()


class EventsWrapper(EventWrapper):
    """
    A class that wrap event/command callbacks.
    It inherits from `EventWrapper` in order to use `self` param.
    """

    server = Server(manager)

    # add `on_data_received` coroutine as a callback for the event `on_data_received`.
    # type hinting data with `Data` forces the parsing from the passed parameter to a `Data` type.
    # `Event.before` is an event which is raised before its parent event.
    @server.data_received.before().as_callback()
    async def on_data_received(self, data: Data):
        print(f"Data received : {data!r}.")

    @server.data_received.as_callback()
    async def process_data(self, data: Data):
        # simulate processing time
        await asyncio.sleep(1)

        # random choice between lose or not data
        if round(random.random()) == 0:
            # raises the data lost event if the data are lost
            await self.server.data_lost.raise_event()

        else:
            # raises the data lost event if the data are OK
            await self.server.data_ok.raise_event()

    @server.data_lost.as_callback()
    async def on_data_lose(self):
        raise RuntimeError("Data lost")

    @server.data_ok.as_callback()
    async def on_data_ok(self):
        print("Data OK !")

    # uses `.after` event which refers to an event that is raised,
    # the first argument is the time took by the previous event's callbacks
    # when all if the parent event's callbacks finish
    @server.data_received.after(pass_extra=True).as_callback()
    async def after_data_received(self, time_took: int, data: Data):
        print(f"Processing {data} took {time_took} seconds")

    # before any parent's event (shutdown) is invoked
    @server.shutdown.before().as_callback()
    async def shutting_down(self):
        print("Shutting down...")

        # simulate shutting down processing time
        await asyncio.sleep(3)

    @server.shutdown.as_callback()
    async def ready_to_shutdown(self):
        print("Server shut down.")


# the manager error handler event.
# It is raised when a error occur in a manager's event only if the param `handle_errors` is set to true (default).
# this event pass in the parameters the error, the event and the callback where the error happened.
# args and kwargs are the arguments/keyword arguments passed when the event has be raised.
@manager.error_handler.as_callback()
async def on_error(error: Exception, event: Event, callback: Callback, *args, **kwargs):
    warn(
        Warning(
            f"An {type(error)} error occurred while raising {event.event_name!r} event -> Callback : {callback.__name__!r}."
            f"\nPassed args : {args, kwargs}."
            f"\n {error}"
        )
    )


async def main():
    # raises the start event
    await start_event.raise_event("A cool data")


if __name__ == "__main__":
    events = EventsWrapper()

    # displays the registered events and commands
    print(f"Registered events : {[e.event_name for e in manager.events]}")
    print(f"Registered commands : {[e.command_name for e in manager.commands]}.")

    # main
    manager.loop.run_until_complete(main())



import asyncio
from asyevent import EventManager


# creates a main event manager
manager = EventManager()

# creates an event
sample_event = manager.create_event("sample_event")


# adds `call_on_event` coroutine as sample_event's callback
@sample_event.as_callback()
async def sample_event_callback(text: str):
    print(text)
    await asyncio.sleep(2)


# uses `.after` event which refers to an event that is raised
# when the parent event's callbacks are ended (data_lost)
@sample_event.after(pass_extra=True).as_callback()
async def after_event(time_took: int, *args):
    # invokes the command `say`
    await manager.invoke_command("say", f"I've been here for {time_took} seconds")


# adds the `say_stm` coroutine as a callback of the command `say`
@manager.as_command(name="say")
async def say_stm(name: str):
    print(f"Hello, {name} !")


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    # raises the event `sample_event`
    loop.run_until_complete(sample_event("Hello, world !"))



from __future__ import annotations

import asyncio
import inspect

from typing import Callable, Union, Tuple, Optional
from asyevent.utils.parser import parse_parameters


class Callback:
    """
    Callable objet, wraps the coroutine behavior with options and errors handler.
    """

    def __init__(
        self,
        callback: Union[Callable, Callback],
        *,
        is_active: bool = True,
        refuse_handling: bool = False,
        continue_on_error: bool = False,
        loop: int = 1,
        loop_delay: float = 0,
        start_delay: float = 0,
        checker: Callable[[tuple, dict], bool] = None,
    ):
        """
        Initialises a callback with a coroutine and options.

        :param callback: The coroutine that will be executed on invoke.
        :param is_active: Is the coroutine active.
        :param refuse_handling: Even if an handler is pass in `.invoke()` method,
            exceptions will be raised if it is set to `True`.

        :param continue_on_error: When a exception occur, if it's handled,
            defines if the loop iterations should continue.

        :param loop: The coroutine will be called `loop` times.
        :param loop_delay: The delay in seconds between each loop iteration.
        :param start_delay: The delay in seconds before the coroutine call.
            It does not impact the event raising process.

        :param checker: A lambda that takes as parameters the callbacks ones and returns a boolean.
            If it returns `False`, the call is aborted.
            `checker` is called at each iteration. If an error occurred, it is handled from the error handler.

        :raise TypeError: If the `coroutine` parameter is not a coroutine.:
        """
        if isinstance(callback, Callback):
            # copy old callback's attributes.
            self.__dict__ = callback.__dict__

            return

        self._coroutine = callback
        self.__name__ = self._coroutine.__name__

        self.is_classmethod = self.signature.parameters.get("self") is not None

        if not inspect.iscoroutinefunction(self._coroutine):
            raise TypeError(
                f"Callback function {self.__name__!r} must be a _coroutine."
            )

        self.is_running = False
        self.wrapper = None

        self.is_active = is_active
        self.refuse_handling = refuse_handling
        self.continue_on_error = continue_on_error

        self.loop = loop
        self.loop_delay = loop_delay
        self.start_delay = start_delay

        self.checker = checker

    async def __call__(self, *args, _event=None, _handler=None, **kwargs):
        """
        Invokes the coroutine.

        :param _event: If an error is handled, the event will be passed as a parameter.
        :param _handler: If an error occur, the `_handler` event will be raised.
        :param args: Coroutine parameters.
        :param kwargs: Coroutine keyword parameters.

        :raise Exception: If an exception cannot be handle, it will be raised.:
        """
        await self.invoke(*args, **kwargs, _event=_event, _handler=_handler)

    @property
    def signature(self) -> inspect.Signature:
        """
        The coroutine signature.

        :return: The coroutine signature.
        """
        return inspect.signature(self._coroutine)

    async def invoke(self, *args, _event=None, _handler=None, **kwargs):
        """
        Invokes the coroutine.

        :param _event: If an error is handled, the event will be passed as a parameter.
        :param _handler: If an error occur, the `_handler` event will be raised.
        :param args: Coroutine parameters.
        :param kwargs: Coroutine keyword parameters.

        :raise Exception: If an exception cannot be handle, it will be raised.:
        """
        if self.is_active:
            self.is_running = True
            await asyncio.sleep(self.start_delay)

            async def iterate():
                # stop iteration if an error is returned.
                if isinstance(
                    await self._invoke_once(
                        *args,
                        **kwargs,
                        _event=_event,
                        _handler=_handler,
                    ),
                    Exception,
                ):
                    return

            if self.loop <= 0:
                while True:
                    await iterate()

            else:
                for _ in range(self.loop):
                    await iterate()

            self.is_running = False

    async def _invoke_once(
        self, *args, _event=None, _handler=None, **kwargs
    ) -> Optional[Exception]:
        a, kwa = self._parse_arguments(*args, **kwargs)

        if self.checker is None or self.checker(*a, **kwa):
            try:
                await self._coroutine(*a, **kwa)

            except Exception as e:
                if _handler is not None and not self.refuse_handling:
                    asyncio.create_task(
                        _handler.raise_event(e, _event, self, *args, **kwargs)
                    )

                    if not self.continue_on_error:
                        return e

                else:
                    raise e

        if self.loop > 1:
            await asyncio.sleep(self.loop_delay)

    def _parse_arguments(self, *args, **kwargs) -> Tuple[tuple, dict]:
        args, kwargs = parse_parameters(self._coroutine, *args, **kwargs)

        if self.is_classmethod:
            if self.wrapper is None:
                raise ValueError(
                    f"Missing wrapper instance on classmethod callback {self.__name__!r}. "
                    f"\n Commune causes are : "
                    f"\n * Missing to to call EventWrapper.init_callbacks()."
                    f"\n * Declaring callback as class callbacks."
                    f"\n * Overriding this callback method."
                )

            args = (self.wrapper,) + tuple(args)

        return args, kwargs

    def enable(self):
        """
        Enables the callback.
        """
        self.is_active = True

    def disable(self):
        """
        Disables the callback.
        """
        self.is_active = False

    def loop(self, *, times: int, delay: float = 0):
        """
        Shortcut for loop settings.

        :param times: How many times the _coroutine iterate on invocation.
        :param delay: Delay between _coroutine iterations.
        """
        self.loop = times
        self.loop_delay = delay



from __future__ import annotations

import typing

from inspect import signature

from abc import abstractmethod

from asyevent.exceptions import ParsingError, ParsingNotImplemented

# the list of builtins types that can be use for parsing values
__parsable_builtins__ = {str, int, float, dict, list, tuple, set, frozenset}


@typing.runtime_checkable
class _IParsable(typing.Protocol):
    """
    Parser protocol that provides a parser method call when passing parameters to callbacks.
    """

    @classmethod
    @abstractmethod
    def __parse__(cls, value: typing.Any) -> _IParsable:
        """
        Parsing method to override.
        Raise `asyevent.exceptions.ParsingError` for parsing errors.

        Type hinting forces a match before parsing.
        Otherwise, `ParsingError` is raised.
        """
        pass


def _parse_parameter(cls: type, value: typing.Any) -> typing.Any:
    if typing.get_origin(cls) is typing.Union:
        cls = typing.get_args(cls)[0]

    if cls is typing.Any or cls is None or isinstance(value, cls):
        return value

    if cls in __parsable_builtins__:
        try:
            return cls(value)

        except ValueError:
            raise ParsingError(value=value, excepted_type=cls)

    elif issubclass(cls, _IParsable):
        params = [p.annotation for p in signature(cls.__parse__).parameters.values()]

        # checks if the values passed matches the parsing method
        if type(params[0]) is not type(value):
            raise ParsingError(value=value, excepted_type=cls)

        return cls.__parse__(value)

    raise ParsingNotImplemented(value=value, excepted_type=cls)


def parse_parameters(f: typing.Callable, *args, **kwargs) -> typing.Tuple[tuple, dict]:
    """
    Parse parameters to match with the signature.
    Parsing to `typing.Union` will try to parse into the first possibility.

    :raise `ParsingError`: If a parameter cannot be parsed while it is type hinted.
    :return:
    """
    types = {k: typing.Any for k in dict(signature(f).parameters).keys()}
    types.pop("self", None)

    # merging signatures names and hint types
    hints: dict[str, typing.Any] = types | typing.get_type_hints(f)

    kwargs = {k: _parse_parameter(hints.get(k), v) for k, v in kwargs.items()}

    delta = max(0, len(args) - len(hints.values()))
    final_hints = list(hints.values()) + [typing.Any] * delta

    args = tuple(_parse_parameter(final_hints[i], arg) for i, arg in enumerate(args))

    return args, kwargs



"""
An implementation of events and asynchronous callbacks using decorators.

Code documentation is not currently complete.
Any contribution is greatly appreciated.

example :
    import asyncio

    manager = EventManager()
    sample_event = manager.create_event('sample_event')


    @sample_event.as_callback()
    async def call_on_event(text: str):
        print(text)

    manager.loop.run_until_complete(sample_event('Hello, world !'))    # could use `Event`.raise_event() instead.

"""

from asyevent.event import Event
from asyevent.event_manager import EventManager
from asyevent.event_wrapper import EventWrapper
from asyevent.callback import Callback
from asyevent.command import Command


__author__ = "Zucchinetti Hervé"
__status__ = "In development"
__version__ = "0.2.8"



from __future__ import annotations

from asyevent.callback import Callback
from asyevent.event import Event
from asyevent.exceptions import CommandAlreadyExists

from typing import Callable, Union


class Command(Event):
    """
    Do not manually initialise this class. Use instead `EventManager.create_command()`.

    A `Command` object is not different of a `Event` one.
    Commands are stored separately from events and are created with an initial callback,
    that is making easier an usage of one callback for on event.
    """

    _COMMAND_EVENT_FORMAT = "<command:{}>"

    def __init__(
        self,
        callback: Union[Callable, Callback],
        *,
        event_manager,
        name: str = None,
        handle_errors: bool = True,
        multiple_callbacks: bool = True,
        priority: int = 1,
        **options
    ):
        """
        Initialises a command with an initial coroutine or callback and event parameters.

        :param coroutine: The initial coroutine or callback from what the command is created.
        :param event_manager: An event manager
        :param name: The command name. Must be unique in the event manager.
        :param handle_errors: Do the errors handled into the `event_manager` error handler.
        :param multiple_callbacks: Does the command allow multiple callbacks associated to.
        :param priority: When the command will be raised, the callbacks are invoked int a priority ascending order.
        :param options: The callback options.

        :raise CommandAlreadyExists: If the command name is not unique in the event_manager.
        """
        self.command_name = name or callback.__name__

        if event_manager.get_command(self.command_name, case_sensitive=False):
            raise CommandAlreadyExists(event_manager.get_command(self.command_name))

        super().__init__(
            name=self._COMMAND_EVENT_FORMAT.format(self.command_name),
            event_manager=event_manager,
            handle_errors=handle_errors,
            multiple_callbacks=multiple_callbacks,
        )

        self._initial_callback = self.create_callback(
            callback, priority=priority, **options
        )

    @property
    def initial_callback(self) -> Callback:
        """
        The callback from what the command was created.

        :return: The initial callback.
        """
        return self._initial_callback
